// ----------------------------------------------------------------------------
// ILSS_PAR 
// ----------------------------------------------------------------------------
// Parallel Interval Linear System Solver
// Approximate matrix inversion module
// ----------------------------------------------------------------------------
// Paralleler Verifizierter Loeser fuer lineare Intervallgleichungssysteme
// Modul zur naeherungsweisen Invertierung einer Matrix
// ----------------------------------------------------------------------------
// Copyright (C) 2004-2007 Markus Grimmer
//
// markus.grimmer@math.uni-wuppertal.de
// ----------------------------------------------------------------------------

// This file is part of the Parallel Interval Linear System Solver ILSS_PAR.
// 
// ILSS_PAR is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// ILSS_PAR is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with ILSS_PAR; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

/**
 * @file matinv_aprx_par.cpp
 * 
 * ILSS_PAR Parallel Interval Linear System Solver 
 * Approximate matrix inversion module.
 * 
 * Copyright (C) 2004-2007 Markus Grimmer
 *
 * Parallel approximate matrix inversion using LU decomposition
 * and forward-backward substitution,
 * column distribution, pivoting and send ahead
 */

#include <vector>
#include <iterator>
#include <algorithm>

#include "cxsc_mpicomm_templ.hpp"   // MPI Communication for C-XSC
#include "vectoroperations.hpp"     // additional ops for STL vectors
#include "matinv_aprx_par.hpp"      // Local header

using namespace cxsc;
using namespace std;


// ----------------------------------------------------------------------------

/**
 * Distributes columns among the involved processes.
 * 
 * @param[in]   A         The input matrix
 * @param[out]  MyA       Distributed matrix columns
 * @param[out]  mycolumns Local column subset
 * @param[in]   lb1       Matrix index bound
 * @param[in]   ub1       Matrix index bound
 * @param[in]   lb2       Matrix index bound
 * @param[in]   ub2       Matrix index bound
 * @param[out]  mycol     Local to global column mapping
 * @param[in]  procs      Number of parallel processes
 * @param[in]  mypid      Process ID
 * @param[out] commerrc   Error code
 */
void ColumnDistribution(rmatrix& A, rmatrix& MyA, vector<int>& mycolumns,
                        int lb1, int ub1, int lb2, int ub2,
                        intvector& mycol, int procs, int mypid, 
                        int& commerrc)
{

  MPI_Status status;

  Resize(mycol,lb2,ub2);

  for(int i = lb2; i <= ub2; i++) 
  {
    mycol[i]=0;
  }

  for (int i=lb2; i<=ub2; i++)
  {
    if (((i-lb2)%procs)==mypid)
    {
      mycolumns.push_back(i);
      mycol[i]=mycolumns.size();
    } 
  }

  MyA=rmatrix(lb1,ub1,1,mycolumns.size());
  MyA=0.0;

  for (int i=lb2; i<=ub2; i++)
  {
    if (mypid==0)
    {
      // Send row from 0 to (i-lb2)%procs
      if ((i-lb2)%procs!=0)
      {
        commerrc=MPI_Send(A, lb1, ub1, i, i, (i-lb2)%procs, 0, MPI_COMM_WORLD);
      }
      else 
      {
        MyA[Col(mycol[i])]=A[Col(i)];
      }
    }
    else if (mypid==((i-lb2)%procs))
    {
      commerrc==MPI_Recv(MyA, lb1, ub1, mycol[i], mycol[i], 0, 
                     0, MPI_COMM_WORLD, &status); 
    }
  }

} //ColumnDistribution


// ----------------------------------------------------------------------------

/**
 * Matrix inversion function.
 * 
 * This version: 
 * Includes parameters for already distributed data 
 * (matrix columns) and already initialized parallel environment
 * 
 * @param[in,out] A         The input matrix (Remark: which will be overwritten
                            with the LU factorization)
 * @param[out]    R         Approximate inverse
 * @param[in]  MyA        Distributed matrix columns
 * @param[in]  mycolumns  Local column subset
 * @param[in]  mycol      Local to global column mapping
 * @param[in]  distribute Distribution of columns necessary?
 * @param[in]  procs      Number of parallel processes
 * @param[in]  mypid      Process ID
 * @param[out] errc       Error code
 * @param[out] commerrc   Error code
 * @param[in,out] ausg    Output file stream
 */
void MatInv(rmatrix& A, rmatrix& R, 
            rmatrix& MyA, vector<int>& mycolumns, intvector& mycol, 
            bool distribute,
            int procs, int mypid,
            int& errc, int& commerrc, ofstream& ausg) 

{
  //----------------------------------------------------------------------------
  // MPI preparation
  //----------------------------------------------------------------------------

  MPI_Status status;

  int   lb1, lb2, ub1, ub2, piv, ptmp;  
  const real eps = 1e-15;
  bool ok=true;

  if (mypid==0)
  {
    lb1 = Lb(A,ROW); ub1 = Ub(A,ROW);
    lb2 = Lb(A,COL); ub2 = Ub(A,COL);
    commerrc=MPI_Bcast(&lb1,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&ub1,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&lb2,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&ub2,1,MPI_INT,0,MPI_COMM_WORLD);
  }
  else
  {
    commerrc=MPI_Bcast(&lb1,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&ub1,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&lb2,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&ub2,1,MPI_INT,0,MPI_COMM_WORLD);
    Resize(A,lb1,ub1,lb2,ub2);
  }

  Resize(R,lb1,ub1,lb2,ub2);
  
  intvector pv(lb1,ub1);
  rmatrix L(lb1,ub1,1,1);
  
  ok = ( ((ub1-lb1)==(ub2-lb2))&&(lb1==lb2) )? true : false; // square matrix ?

  // init pivot and L vectors
  for(int i = lb1; i <= ub1; i++) 
  { 
    pv[i] = i;
    L[i][1] = 0.0;
  }   

  //----------------------------------------------------------------------------
  // Column distribution:
  //----------------------------------------------------------------------------
  
  if (distribute)
  {
    ColumnDistribution(A,MyA,mycolumns,lb1,ub1,lb2,ub2,mycol,procs,mypid,commerrc);
  }
  
  //----------------------------------------------------------------------------
  // LU decomposition
  //----------------------------------------------------------------------------

  ausg << "Starte LU-Zerlegung..." << endl;
  
  //----------------------------------------------------------------------------
  // Start decomposition algorithm
  //----------------------------------------------------------------------------

  if (find(mycolumns.begin(),mycolumns.end(),lb2)!=mycolumns.end())
  // ~"if (lb2 in mycolumns)"
  {
     // ---------------
     // Pivot search #1
     // ---------------

     piv=lb1;
     for (int i=lb1+1; i<=ub1; i++)
     {            
        if (abs(MyA[i][mycol[lb2]])>abs(MyA[piv][mycol[lb2]])) 
          piv=i;  
     }            
                  
     // -------------------
     // Update pivot column
     // -------------------
                  
     if (MyA[piv][mycol[lb2]]==0)
       ok=false;
     else
     {
       for (int m=lb1; m<=ub1; m++)
       {            
          if (pv[m]!=piv)
          {
            MyA[pv[m]][mycol[lb2]] = MyA[pv[m]][mycol[lb2]]/MyA[piv][mycol[lb2]];
          }
       }            
       L[Col(1)]=MyA[Col(mycol[lb2])]; //Store the column locally...
     }             
     // --------------------
     // Broadcast "Send" #1 
     // --------------------

     commerrc=MPI_Bcast(&piv,1,MPI_INT,0,MPI_COMM_WORLD);
     commerrc=MPI_Bcast(MyA,lb1,ub1,mycol[lb2],mycol[lb2],0,MPI_COMM_WORLD); 
     
  } //end if lb2 in mycolumns
  
  // ---------------
  // Main loop
  // ---------------
           
  for (int k=lb2; k<=ub2-1; k++)  // for cols
  {
    // -----------------------
    // Broadcast "Receive" #k
    // -----------------------

    if (find(mycolumns.begin(),mycolumns.end(),k)==mycolumns.end())
    {
      commerrc=MPI_Bcast(&piv,1,MPI_INT,(k-lb2)%procs,MPI_COMM_WORLD);
      commerrc=MPI_Bcast(L,lb1,ub1,1,1,(k-lb2)%procs,MPI_COMM_WORLD); 
    }
    else
    {
      L[Col(1)]=MyA[Col(mycol[k])];
    }

    // ----------------------------
    // Store the result (everyone):
    // ----------------------------

    A[Col(k)]=L[Col(1)]; // Overwrite A with factorized columns...

    // --------------------
    // Swap index entries
    // --------------------

    ptmp = pv[piv]; pv[piv] = pv[k]; pv[k] = ptmp;

    // ------------------------
    // Transform rest of matrix
    // -----------------------

    for (int j=k+1; j<=ub2; j++)  // for columns
    {
      // ----------------------
      // Update column if owner
      // ----------------------

      if (find(mycolumns.begin(),mycolumns.end(),j)!=mycolumns.end())
      { //"if j in mycolumns"
      
        for (int i=k+1; i <= ub1; i++)  // for rows
        {
          real h = MyA[pv[i]][mycol[j]]-L[pv[i]][1]*MyA[pv[k]][mycol[j]];  

          if( h == 0.0) // replace 0 by (old element)*eps for numerical stability reasons
            MyA[pv[i]][mycol[j]] *= eps;
          else 
            MyA[pv[i]][mycol[j]] = h;
        }

        // ----------------------
        // Determine new pivot
        // ----------------------

        if ((j==(k+1)) && (j<ub2)) 
        {
          piv=k+1;

          for (int i=k+1; i<=ub1; i++)
          {          
             if (abs(MyA[pv[i]][mycol[k+1]])>abs(MyA[pv[piv]][mycol[k+1]])) 
               piv=i;  
          }            

          // ----------------------
          // Update pivot column
          // ----------------------
          
          if (MyA[pv[piv]][mycol[k+1]]==0)
          {
            ok=false;
          }
          else             
          {
            for (int m=k+1; m<=ub1; m++)
            {  
               if (m!=piv)
               {
                 MyA[pv[m]][mycol[k+1]] = MyA[pv[m]][mycol[k+1]]/MyA[pv[piv]][mycol[k+1]];
               }
            }            
                       
            commerrc=MPI_Bcast(&piv,1,MPI_INT,
                               (k+1-lb2)%procs,MPI_COMM_WORLD);
            commerrc=MPI_Bcast(MyA,lb1,ub1,mycol[k+1],mycol[k+1],
                               (k+1-lb2)%procs,MPI_COMM_WORLD); 
          } // end if (A[...]) else
        } //end if j==

      }//end if j in mycolumns

    } // end for j
  } // end for k


  //----------------------------------------------------
  // Broadcast last column (which is missing until now):
  //----------------------------------------------------
  
  if (((ub2-lb2)%procs)==mypid)
  { //if "ub2 in mycolumns"
    commerrc=MPI_Bcast(MyA,lb1,ub1,mycol[ub2],mycol[ub2],
                       (ub2-lb2)%procs,MPI_COMM_WORLD);
    A[Col(ub2)]=MyA[Col(mycol[ub2])];
  }
  else
  {
    commerrc=MPI_Bcast(A,lb1,ub1,ub2,ub2,(ub2-lb2)%procs,MPI_COMM_WORLD);
  }

  //----------------------------------------------------------------------------
  // Forward/backward substitution
  //----------------------------------------------------------------------------
  
  int start=lb1, end=ub1;
  rvector col(lb1,ub1);

  ausg << "Loese Dreieckssysteme.." << endl;

  int ncols=ub1-lb1+1;
 
  //-----------------
  // Range assignment
  //-----------------
 
  if (mypid<=ncols%procs) 
    start=lb1+mypid*((ncols/procs)+1);
  else 
    start=lb1+(ncols%procs)*((ncols/procs)+1)+(mypid-(ncols%procs))*(ncols/procs);
  if (mypid<ncols%procs) 
    end=start+ncols/procs;
  else
    end=start+ncols/procs-1;

  for(int i=start; i<=end; i++)
  { 
    for(int j=lb1; j<=ub1; j++) col[j] = 0.0;
    col[i] = 1.0; 

    int ii=0;
    dotprecision sum; 
 
    //--------------------------------------------------------------------------
    // Forward substitution with implicitly permuted matrix
    //--------------------------------------------------------------------------
    
    for(int k=lb1; k<=ub1; k++)
    {
      sum = col[k];
      if(ii){
       	for(int m=ii; m<=k-1;m++) accumulate(sum,-A[pv[k]][m],col[m]); }
      else if(sum != 0.0) ii=k;
      col[k] = rnd(sum);
    }

    //--------------------------------------------------------------------------
    // Backward substitution with implicitly permuted matrix
    //--------------------------------------------------------------------------

    for(int k=ub1; k>=lb1; k--)
    {
      sum = col[k];
      for(int m=k+1; m<=ub1; m++) accumulate(sum,-A[pv[k]][m],col[m]);
      col[k] = rnd(sum) / A[pv[k]][k]; 
    }

    R[Col(pv[i])]=col;    

  }// for i=1;...

  //--------------------------------------------------------------------------
  // Gather result
  //--------------------------------------------------------------------------

  if (mypid==0)
  {
    int istart, iend;

    // for all procs. without myself! :
    for (int i=1; i<procs; i++)  
    {
      if (i<=ncols%procs)
        istart=lb1+i*((ncols/procs)+1);
      else
        istart=lb1+(ncols%procs)*((ncols/procs)+1)+(i-(ncols%procs))*(ncols/procs);
      if (i<ncols%procs)
        iend=istart+ncols/procs;
      else
        iend=istart+ncols/procs-1;
      
      for (int ii=istart; ii<=iend; ii++)
      {
        commerrc=MPI_Recv(R,lb1,ub1,pv[ii],pv[ii], i, 0, MPI_COMM_WORLD, &status);
      }
      
    }
  }
  else
  {
    for (int i=start; i<=end; i++)
      commerrc=MPI_Send(R, lb1, ub1, pv[i], pv[i], 0, 0, MPI_COMM_WORLD);
  }
   
  errc = !ok; 

  if (ok)
    ausg << "Matrix inversion done. " << endl;
  else
    ausg << "Matrix inversion failed. " << endl;
}

// ----------------------------------------------------------------------------

/**
 * Matrix inversion function.
 * 
 * This version: 
 * Does not includes parameters for already distributed data 
 * -> Column distribution inside the function.
 * Includes parameters for already initialized parallel environment.
 * 
 * @param[in,out] A       The input matrix (Remark: which will be overwritten
                          with the LU factorization)
 * @param[out]    R       Approximate inverse
 * @param[in]  procs      Number of parallel processes
 * @param[in]  mypid      Process ID
 * @param[out] errc       Error code
 * @param[out] commerrc   Error code
 * @param[in,out] ausg    Output file stream
 */

void MatInv(rmatrix& A, rmatrix& R, 
            int procs, int mypid, int& errc, int& commerrc, ofstream& ausg) 
{
  vector<int> mycolumns;
  intvector mycol;
  rmatrix MyA;

  MatInv(A,R,MyA,mycolumns,mycol,true,procs,mypid,errc,commerrc,ausg);

} //MatInv (interface)


// ----------------------------------------------------------------------------

/**
 * Matrix inversion function.
 * 
 * This version: 
 * Does not includes parameters for already distributed data or parallel environment
 * -> Column distribution and parallel environment initialization inside the function.
 * 
 * @param[in,out] A       The input matrix (Remark: which will be overwritten
                          with the LU factorization)
 * @param[out]    R       Approximate inverse
 * @param[out] errc       Error code
 * @param[out] commerrc   Error code
 * @param[in,out] ausg    Output file stream
 */

void MatInv(rmatrix& A, rmatrix& R, int& errc, int& commerrc, ofstream& ausg) 
{
  int procs;
  int mypid;

  commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid);
  commerrc=MPI_Comm_size(MPI_COMM_WORLD, &procs);

  vector<int> mycolumns;
  intvector mycol;
  rmatrix MyA;

  MatInv(A,R,MyA,mycolumns,mycol,true,procs,mypid,errc,commerrc,ausg);

} //MatInv (interface)



